# chgsupport - functionality to support cHg
#
# Copyright 2011 Yuya Nishihara <yuya@tcha.org>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

"""functionality to support cHg

hg serve --cmdserver=unix
    communicate with cmdserver via unix domain socket

This extends channels and commands of cmdserver:

's'-channel
    handles util.system() request

'P'-channel
    handles getpass() request

'chdir' command
    change current directory

'getpager' command
    checks if pager is enabled and which pager should be executed

'getpid' command
    get process id of the server (worker)

'setchguiattr' command
    set initial value of ui attributes

'setenv' command
    replace os.environ completely

Configuration::

    [chg]
    maxworkers = 2  # the number of the worker processes which run cmdserver

If `maxworkers = 0`, it serves the worker functionality on the master process.
This mode is only available for debugging purpose.
"""

import errno, os, signal, socket, struct, tempfile
from mercurial import cmdutil, commands, commandserver, dispatch, error, \
                      extensions, i18n, scmutil, util
from mercurial.i18n import _

class channeledinput(commandserver.channeledinput):
    def readpass(self):
        return self._read(self.maxchunksize, 'P').rstrip('\n')

class channeledsystem(object):
    """Request to execute system() in the following format:

    command length (unsigned int),
    cmd

    and waits:

    exitcode length (unsigned int),
    exitcode (int)
    """
    def __init__(self, in_, out, channel):
        self.in_ = in_
        self.out = out
        self.channel = channel

    # copied from mercurial/util.py
    def __call__(self, cmd, environ={}, cwd=None, onerr=None, errprefix=None,
                 out=None):
        # XXX no support for environ, cwd and out

        origcmd = cmd
        cmd = util.quotecommand(cmd)

        self.out.write(struct.pack('>cI', self.channel, len(cmd)))
        self.out.write(cmd)
        self.out.flush()

        length = self.in_.read(4)
        length = struct.unpack('>I', length)[0]
        if length != 4:
            rc = 255
        else:
            rc = struct.unpack('>i', self.in_.read(4))[0]

        if rc and onerr:
            errmsg = '%s %s' % (os.path.basename(origcmd.split(None, 1)[0]),
                                util.explainexit(rc)[0])
            if errprefix:
                errmsg = '%s: %s' % (errprefix, errmsg)
            try:
                onerr.warn(errmsg + '\n')
            except AttributeError:
                raise onerr(errmsg)
        return rc

class nulldirstate(object):
    pass

class nullrepo(object):
    def __init__(self, baseui):
        self.baseui = baseui
        self.ui = baseui.copy()
        self.dirstate = nulldirstate()

    def invalidate(self):
        pass

    def invalidatedirstate(self):
        pass

    def __nonzero__(self):
        return False

class chgcmdserver(commandserver.server):
    def __init__(self, sui, ui, repo, fin, fout):
        super(chgcmdserver, self).__init__(ui, repo, mode='pipe')
        self._sui = sui  # ui for server output
        self.cerr = commandserver.channeledoutput(fout, fout, 'e')
        self.cout = commandserver.channeledoutput(fout, fout, 'o')
        self.cin = channeledinput(fin, fout, 'I')
        self.cresult = commandserver.channeledoutput(fout, fout, 'r')
        self.csystem = channeledsystem(fin, fout, 's')
        self.client = fin

    def chdir(self):
        """Change current directory

        Note that the behavior of --cwd option is bit different from this.
        It does not affect --config parameter.
        """
        length = struct.unpack('>I', self._read(4))[0]
        if not length:
            return
        path = self._read(length)
        self._sui.debug('chdir to %r\n' % path)
        os.chdir(path)

    def getpager(self):
        """Read cmdargs and write pager command to r-channel if enabled

        If pager isn't enabled, this writes '\0' because channeledoutput
        does not allow to write empty data.
        """
        length = struct.unpack('>I', self._read(4))[0]
        if not length:
            args = []
        else:
            args = self._read(length).split('\0')

        try:
            pagermod = extensions.find('pager')
        except KeyError:
            self.cresult.write('\0')
            return

        pagercmd = self.ui.config('pager', 'pager', os.environ.get('PAGER'))
        if not pagercmd:
            self.cresult.write('\0')
            return

        try:
            cmd, _func, args, options, _cmdoptions = dispatch._parse(self.ui,
                                                                     args)
        except (error.AmbiguousCommand, error.CommandError,
                error.UnknownCommand):
            self.cresult.write('\0')
            return

        # duplicated from hgext/pager.py
        attend = self.ui.configlist('pager', 'attend', pagermod.attended)
        auto = options['pager'] == 'auto'
        always = util.parsebool(options['pager'])
        if (always or auto and
            (cmd in attend or
             (cmd not in self.ui.configlist('pager', 'ignore')
              and not attend))):
            self.cresult.write(pagercmd)
        else:
            self.cresult.write('\0')

    def getpid(self):
        """Write pid of the server (worker)"""
        self.cresult.write(struct.pack('>i', os.getpid()))

    def setchguiattr(self):
        """Set ui attributes if not configured"""
        length = struct.unpack('>I', self._read(4))[0]
        if not length:
            return
        s = self._read(length)
        for l in s.splitlines():
            k, v = l.split(': ', 1)
            if self.ui.config('ui', k) is None:
                self._sui.debug('set ui %r: %r\n' % (k, v))
                self.ui.setconfig('ui', k, v)

        # reload config so that ui.plain() takes effect
        for f in scmutil.rcpath():
            self.ui.readconfig(f, trust=True)

        # skip initial uisetup() by 'mode=chgauto'
        if self.ui.config('color', 'mode') == 'chgauto':
            self.ui.setconfig('color', 'mode', 'auto')

    def setenv(self):
        """Clear and update os.environ

        Note that not all variables can make an effect on the running process.
        """
        length = struct.unpack('>I', self._read(4))[0]
        if not length:
            return
        s = self._read(length)
        try:
            newenv = dict(l.split('=', 1) for l in s.split('\0'))
        except ValueError:
            raise ValueError('unexpected value in setenv request')

        i18nchanged = util.any(newenv.get(k) != os.environ.get(k)
                               for k in ('LANG', 'LANGUAGE', 'LC_MESSAGES'))

        if self._sui.debugflag:
            for k in sorted(set(os.environ.keys() + newenv.keys())):
                ov, nv = os.environ.get(k), newenv.get(k)
                if ov != nv:
                    self._sui.debug('change env %r: %r -> %r\n' % (k, ov, nv))
        os.environ.clear()
        os.environ.update(newenv)

        if i18nchanged:
            self._sui.debug('reload i18n module\n')
            reload(i18n)  # recreate gettext object 't'

    capabilities = commandserver.server.capabilities.copy()
    capabilities.update({'chdir': chdir,
                         'getpager': getpager,
                         'getpid': getpid,
                         'setchguiattr': setchguiattr,
                         'setenv': setenv})

def _wrapui(ui):
    class chgui(ui.__class__):
        csystem = None  # should be assigned later

        def copy(self):
            o = super(chgui, self).copy()
            o.csystem = self.csystem
            return o

        def termwidth(self):
            n = self.configint('ui', '_termwidth')
            if n and n > 0:
                return n
            else:
                return super(chgui, self).termwidth()

        # copied from mercurial/ui.py
        def edit(self, text, user):
            (fd, name) = tempfile.mkstemp(prefix="hg-editor-", suffix=".txt",
                                          text=True)
            try:
                f = os.fdopen(fd, "w")
                f.write(text)
                f.close()

                editor = self.geteditor()

                self.csystem("%s \"%s\"" % (editor, name),
                             environ={'HGUSER': user},
                             onerr=util.Abort, errprefix=_("edit failed"),
                             out=self.fout)

                f = open(name)
                t = f.read()
                f.close()
            finally:
                os.unlink(name)

            return t

        # copied from mercurial/ui.py
        def getpass(self, prompt=None, default=None):
            if not self.interactive():
                return default
            try:
                self.fout.write(prompt or _('password: '))
                self.fout.flush()
                return self.fin.readpass()
            except EOFError:
                raise util.Abort(_('response expected'))

        # copied from mercurial/ui.py
        def plain(self, feature=None):
            plain = self.configbool('ui', '_plain')
            plainexcept = self.configlist('ui', '_plainexcept')
            if not plain and not plainexcept:
                return False
            if feature and plainexcept:
                return feature not in plainexcept
            return True

    ui.__class__ = chgui

class TaintedWorkerError(error.Abort):
    pass

class service(object):
    """Prefork server for handling connection from cHg client"""
    def __init__(self, ui, sockpath):
        self._ui = ui
        self._sockpath = sockpath
        self._maxworkers = self._ui.configint('chg', 'maxworkers', 2)

    def init(self):
        util.setsignalhandler()
        self._sock = socket.socket(socket.AF_UNIX)
        self._sock.bind(self._sockpath)

    def run(self):
        self._sock.listen(1)
        self._ui.note(_('listening at %s (pid=%d)\n')
                      % (self._sockpath, os.getpid()))

        self._workerpids = []
        master = self._maxworkers > 0
        try:
            try:
                while master:
                    master = self._servemaster()
                if not master:
                    self._initworkerui()
                    while True:
                        self._serveworker()
            except error.SignalInterrupt:
                if master:
                    for pid in self._workerpids:
                        self._ui.debug('terminate worker process (pid=%d)\n'
                                       % pid)
                        os.kill(pid, signal.SIGTERM)
        finally:
            self._sock.close()
            if master or self._maxworkers <= 0:
                os.unlink(self._sockpath)

    def _servemaster(self):
        for _i in xrange(self._maxworkers - len(self._workerpids)):
            pid = os.fork()
            if pid == 0:
                return False
            self._ui.debug('forked worker process (pid=%d)\n' % pid)
            self._workerpids.append(pid)

        pid, _status = os.wait()
        self._workerpids.remove(pid)
        return True

    def _initworkerui(self):
        self._workerui = self._ui.copy()
        # don't retain command-line opts of the service
        for opt in ('verbose', 'debug', 'quiet', 'traceback'):
            self._workerui.setconfig('ui', opt, None)
        self._workerui.fixconfig()
        _wrapui(self._workerui)

    def _serveworker(self):
        conn, _addr = self._sock.accept()
        try:
            self._ui.note('accepted connection (pid=%d)\n' % os.getpid())
            try:
                fin = conn.makefile('rb')
                fout = conn.makefile('wb')
                try:
                    self._runcmdserver(fin, fout)
                finally:
                    fin.close()
                    # implicit flush() may cause another EPIPE
                    fout.close()
            except IOError, err:
                if err.errno != errno.EPIPE:
                    raise
        finally:
            conn.close()

    def _runcmdserver(self, fin, fout):
        ui = self._workerui.copy()
        sv = chgcmdserver(self._ui, ui, nullrepo(ui), fin, fout)
        ui.csystem = sv.csystem

        origcwd = os.getcwd()
        origcmdcnt = len(commands.table)
        origextcnt = len(extensions._extensions)
        try:
            try:
                sv.serve()
                self._ui.debug('finished (pid=%d)\n' % os.getpid())
            except IOError, err:
                if err.errno == errno.EPIPE:
                    self._ui.debug('finished by broken pipe (pid=%d)\n'
                                   % os.getpid())
                else:
                    raise
        finally:
            os.chdir(origcwd)
        if origcmdcnt != len(commands.table):
            raise TaintedWorkerError('loaded local command or alias')
        if origextcnt != len(extensions._extensions):
            raise TaintedWorkerError('loaded local extension')
        if _extensionsloadedbypath:
            raise TaintedWorkerError('loaded unwanted extension')

def _disablepager():
    try:
        pagermod = extensions.find('pager')
    except KeyError:
        return
    # _runpager(p) or _runpager(ui, p)
    extensions.wrapfunction(pagermod, '_runpager', lambda orig, *args: None)

# extensions.disabledcmd() uses .loadpath(), but .loadpath() does not
# remember loaded extension in _extensions list. Since bad-mannered extension
# may mangle global object, we need to restart cmdserver process even if
# uisetup() or extsetup() is not called.
_extensionsloadedbypath = False
def _extensionsloadpath(orig, path, module_name):
    global _extensionsloadedbypath
    ext = orig(path, module_name)
    _extensionsloadedbypath = True
    return ext

def _serve(orig, ui, repo, **opts):
    mode = opts['cmdserver']
    if not mode or mode != 'unix':
        return orig(ui, repo, **opts)

    _disablepager()
    extensions.wrapfunction(extensions, 'loadpath', _extensionsloadpath)

    # it might be confusing to reuse --port option
    sockpath = opts['port']
    if not sockpath:
        raise util.Abort(_('no socket path specified with --port'))

    svc = service(ui, sockpath)
    cmdutil.service(opts, initfn=svc.init, runfn=svc.run)

def uisetup(ui):
    extensions.wrapcommand(commands.table, 'serve', _serve)
